import io
import json
import os
import pathlib
import re
import shutil
import zipfile
from unittest import mock

import pytest
import responses
import yaml
from pydantic.v1 import ValidationError

from cumulusci.core.config import BaseProjectConfig, TaskConfig, UniversalConfig
from cumulusci.core.dependencies.dependencies import (
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
)
from cumulusci.core.dependencies.github import UnmanagedGitHubRefDependency
from cumulusci.core.exceptions import (
    CumulusCIUsageError,
    DependencyLookupError,
    GithubException,
    PackageUploadFailure,
    TaskOptionsError,
    VcsException,
)
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.salesforce_api.package_zip import BasePackageZipBuilder
from cumulusci.tasks.create_package_version import (
    PERSISTENT_ORG_ERROR,
    CreatePackageVersion,
    PackageConfig,
    PackageTypeEnum,
    VersionTypeEnum,
)
from cumulusci.tests.util import CURRENT_SF_API_VERSION
from cumulusci.utils import temporary_dir, touch

print(CURRENT_SF_API_VERSION)


@pytest.fixture
def repo_root():
    with temporary_dir() as path:
        os.mkdir(".git")
        os.mkdir("src")
        pathlib.Path(path, "src", "package.xml").write_text(
            '<?xml version="1.0" encoding="utf-8"?>\n<Package xmlns="http://soap.sforce.com/2006/04/metadata"></Package>'
        )
        with open("cumulusci.yml", "w") as f:
            yaml.dump(
                {
                    "project": {
                        "dependencies": [
                            {"namespace": "pub", "version": "1.5"},
                            {
                                "repo_owner": "SalesforceFoundation",
                                "repo_name": "EDA",
                                "ref": "aaaaa",
                                "subfolder": "unpackaged/pre/first",
                            },
                            {
                                "namespace": "hed",
                                "version": "1.99",
                            },
                        ]
                    }
                },
                f,
            )
        pathlib.Path(path, "unpackaged", "pre", "first").mkdir(parents=True)
        touch(os.path.join("unpackaged", "pre", "first", "package.xml"))
        yield path


@pytest.fixture
def project_config(repo_root):
    project_config = BaseProjectConfig(
        UniversalConfig(),
        repo_info={"root": repo_root, "branch": "main"},
    )
    project_config.config["project"]["package"]["install_class"] = "Install"
    project_config.config["project"]["package"]["uninstall_class"] = "Uninstall"
    project_config.keychain = BaseProjectKeychain(project_config, key=None)
    pathlib.Path(repo_root, "orgs").mkdir()
    pathlib.Path(repo_root, "orgs", "scratch_def.json").write_text(
        json.dumps(
            {
                "edition": "Developer",
                "settings": {},
            }
        )
    )

    return project_config


@pytest.fixture
def get_task(project_config, devhub_config, org_config):
    def _get_task(options=None):
        opts = options or {
            "package_type": "Managed",
            "org_dependent": False,
            "package_name": "Test Package",
            "static_resource_path": "static-resources",
            "ancestor_id": "04t000000000000",
            "create_unlocked_dependency_packages": True,
            "install_key": "foo",
        }
        task = CreatePackageVersion(
            project_config,
            TaskConfig({"options": opts}),
            org_config,
        )
        with mock.patch(
            "cumulusci.tasks.create_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task._init_task()
        return task

    return _get_task


@pytest.fixture
def task(get_task):
    return get_task()


@pytest.fixture
def mock_download_extract_github():
    with mock.patch(
        "cumulusci.core.dependencies.base.UnmanagedVcsDependency._get_zip_src"
    ) as download_extract_github:
        yield download_extract_github


@pytest.fixture
def mock_get_static_dependencies():
    with mock.patch(
        "cumulusci.tasks.create_package_version.get_static_dependencies"
    ) as get_static_dependencies:
        get_static_dependencies.return_value = [
            PackageNamespaceVersionDependency(namespace="pub", version="1.5"),
            UnmanagedGitHubRefDependency(
                repo_owner="SalesforceFoundation",
                repo_name="EDA",
                subfolder="unpackaged/pre/first",
                ref="abcdef",
            ),
            PackageNamespaceVersionDependency(namespace="hed", version="1.99"),
        ]
        yield get_static_dependencies


class TestPackageConfig:
    def test_org_config(self, project_config, org_config):
        org_config.config_file = None
        with pytest.raises(
            TaskOptionsError,
            match=PERSISTENT_ORG_ERROR,
        ):
            CreatePackageVersion(project_config, TaskConfig(), org_config)

    def test_validate_org_dependent(self):
        with pytest.raises(ValidationError, match="Only unlocked packages"):
            PackageConfig(package_type=PackageTypeEnum.managed, org_dependent=True)  # type: ignore

    def test_validate_post_install_script(self):
        with pytest.raises(ValidationError, match="Only managed packages"):
            PackageConfig(
                package_type=PackageTypeEnum.unlocked, post_install_script="Install"
            )  # type: ignore

    def test_validate_uninstall_script(self):
        with pytest.raises(ValidationError, match="Only managed packages"):
            PackageConfig(
                package_type=PackageTypeEnum.unlocked, uninstall_script="Uninstall"
            )  # type: ignore


class TestCreatePackageVersion:
    devhub_base_url = (
        f"https://devhub.my.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}"
    )
    scratch_base_url = (
        f"https://scratch.my.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}"
    )

    def test_postinstall_script_logic(self, get_task):
        task = get_task({"package_type": "Managed", "package_name": "Foo"})

        # Values set in the fixture project_config above
        assert task.package_config.post_install_script == "Install"
        assert task.package_config.uninstall_script == "Uninstall"

        task = get_task(
            {
                "package_type": "Unlocked",
                "package_name": "Foo",
                "post_install_script": None,
                "uninstall_script": None,
            }
        )

        assert task.package_config.post_install_script is None
        assert task.package_config.uninstall_script is None

    @responses.activate
    def test_run_task(
        self,
        task,
        mock_download_extract_github,
        mock_get_static_dependencies,
        devhub_config,
    ):
        zf = zipfile.ZipFile(io.BytesIO(), "w")
        zf.writestr("unpackaged/pre/first/package.xml", "")
        mock_download_extract_github.return_value = zf
        # _get_or_create_package() responses
        responses.add(  # query to find existing package
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )
        responses.add(  # create Package2
            "POST",
            f"{self.devhub_base_url}/tooling/sobjects/Package2/",
            json={"id": "0Ho6g000000fy4ZCAQ"},
        )

        # _resolve_ancestor_id() responses
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Id": "05i000000000000"}]},
        )

        # _create_version_request() responses
        responses.add(  # query to find existing Package2VersionCreateRequest
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )
        responses.add(  # query to find base version
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "04t000000000002AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "BuildNumber": 1,
                        "IsReleased": False,
                    }
                ],
            },
        )
        responses.add(  # get dependency org API version
            "GET",
            "https://scratch.my.salesforce.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )
        responses.add(  # query for dependency org installed packages
            "GET",
            f"{self.scratch_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "SubscriberPackage": {
                            "Id": "033000000000002AAA",
                            "NamespacePrefix": "pub",
                        },
                        "SubscriberPackageVersionId": "04t000000000002AAA",
                    },
                    {
                        "SubscriberPackage": {
                            "Id": "033000000000003AAA",
                            "NamespacePrefix": "hed",
                        },
                        "SubscriberPackageVersionId": "04t000000000003AAA",
                    },
                ],
            },
        )
        # query dependency org for installed package 1
        responses.add(
            "GET",
            f"{self.scratch_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "04t000000000002AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 5,
                        "PatchVersion": 0,
                        "BuildNumber": 1,
                        "IsBeta": False,
                    }
                ],
            },
        )
        responses.add(  # query dependency org for installed package 2)
            "GET",
            f"{self.scratch_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "04t000000000003AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 99,
                        "PatchVersion": 0,
                        "BuildNumber": 1,
                        "IsBeta": False,
                    }
                ],
            },
        )
        responses.add(  # query for existing package (dependency from github)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {"Id": "0Ho000000000001AAA", "ContainerOptions": "Unlocked"}
                ],
            },
        )
        responses.add(  # query for existing package version (dependency from github)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Id": "08c000000000001AAA"}]},
        )
        responses.add(  # check status of Package2VersionCreateRequest (dependency from github)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "08c000000000001AAA",
                        "Status": "Success",
                        "Package2VersionId": "051000000000001AAA",
                    }
                ],
            },
        )
        responses.add(  # get info from Package2Version (dependency from github)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "SubscriberPackageVersionId": "04t000000000001AAA",
                        "MajorVersion": 0,
                        "MinorVersion": 1,
                        "PatchVersion": 0,
                        "BuildNumber": 1,
                    }
                ],
            },
        )
        responses.add(  # query for existing package (unpackaged/pre)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {"Id": "0Ho000000000004AAA", "ContainerOptions": "Unlocked"}
                ],
            },
        )
        responses.add(  # query for existing package version (unpackaged/pre)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Id": "08c000000000004AAA"}]},
        )
        responses.add(  # check status of Package2VersionCreateRequest (unpackaged/pre)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "08c000000000004AAA",
                        "Status": "Success",
                        "Package2VersionId": "051000000000004AAA",
                    }
                ],
            },
        )
        responses.add(  # get info from Package2Version (unpackaged/pre)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "SubscriberPackageVersionId": "04t000000000004AAA",
                        "MajorVersion": 0,
                        "MinorVersion": 1,
                        "PatchVersion": 0,
                        "BuildNumber": 1,
                    }
                ],
            },
        )
        responses.add(  # create Package2VersionCreateRequest (main package)
            "POST",
            f"{self.devhub_base_url}/tooling/sobjects/Package2VersionCreateRequest/",
            json={"id": "08c000000000002AAA"},
        )
        responses.add(  # check status of Package2VersionCreateRequest (main package)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "08c000000000002AAA",
                        "Status": "Success",
                        "Package2VersionId": "051000000000002AAA",
                    }
                ],
            },
        )
        responses.add(  # get info from Package2Version (main package)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "SubscriberPackageVersionId": "04t000000000002AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "BuildNumber": 1,
                    }
                ],
            },
        )
        responses.add(  # get dependencies from SubscriberPackageVersion (main package)
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Dependencies": {
                            "ids": [
                                {"subscriberPackageVersionId": "04t000000000009AAA"}
                            ]
                        }
                    }
                ],
            },
        )

        with mock.patch(
            "cumulusci.tasks.create_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task()
        zf.close()
        assert task.return_values["dependencies"] == [
            {"version_id": "04t000000000009AAA"}
        ]
        assert task.return_values["install_key"] == task.options["install_key"]
        zf.close()

    @responses.activate
    def test_get_or_create_package__namespaced_existing(
        self, project_config, devhub_config, org_config
    ):
        responses.add(  # query to find existing package
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {"Id": "0Ho6g000000fy4ZCAQ", "ContainerOptions": "Managed"}
                ],
            },
        )

        task = CreatePackageVersion(
            project_config,
            TaskConfig(
                {
                    "options": {
                        "package_type": "Managed",
                        "package_name": "Test Package",
                        "namespace": "ns",
                    }
                }
            ),
            org_config,
        )

        with mock.patch(
            "cumulusci.tasks.create_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task._init_task()

        result = task._get_or_create_package(task.package_config)
        assert result == "0Ho6g000000fy4ZCAQ"

    @responses.activate
    def test_get_or_create_package__exists_but_wrong_type(
        self, project_config, devhub_config, org_config
    ):
        responses.add(  # query to find existing package
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {"Id": "0Ho6g000000fy4ZCAQ", "ContainerOptions": "Unlocked"}
                ],
            },
        )

        task = CreatePackageVersion(
            project_config,
            TaskConfig(
                {
                    "options": {
                        "package_type": "Managed",
                        "package_name": "Test Package",
                        "namespace": "ns",
                    }
                }
            ),
            org_config,
        )
        with mock.patch(
            "cumulusci.tasks.create_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task._init_task()
        with pytest.raises(PackageUploadFailure):
            task._get_or_create_package(task.package_config)

    @responses.activate
    def test_get_or_create_package__devhub_disabled(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json=[{"message": "Object type 'Package2' is not supported"}],
            status=400,
        )

        with pytest.raises(TaskOptionsError):
            task._get_or_create_package(task.package_config)

    @responses.activate
    def test_get_or_create_package__multiple_existing(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 2, "records": []},
        )

        with pytest.raises(TaskOptionsError):
            task._get_or_create_package(task.package_config)

    @responses.activate
    def test_create_version_request__existing_package_version(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Id": "08c000000000001AAA"}]},
        )

        builder = BasePackageZipBuilder()
        result = task._create_version_request(
            "0Ho6g000000fy4ZCAQ", task.package_config, builder
        )
        assert result == "08c000000000001AAA"

    def test_has_1gp_namespace_dependencies__no(self, task):
        assert not task._has_1gp_namespace_dependency([])

    def test_has_1gp_namespace_dependencies__transitive(self, task):
        assert task._has_1gp_namespace_dependency(
            [PackageNamespaceVersionDependency(namespace="foo", version="1.5")]
        )

    def test_convert_project_dependencies__unrecognized_format(self, task):
        with pytest.raises(DependencyLookupError):
            task._convert_project_dependencies([{"foo": "bar"}])

    def test_convert_project_dependencies__no_unlocked_packages(self, task):
        task.options["create_unlocked_dependency_packages"] = False
        assert task._convert_project_dependencies(
            [
                PackageVersionIdDependency(version_id="04t000000000000"),
                UnmanagedGitHubRefDependency(
                    github="https://github.com/test/test", ref="abcdef"
                ),
            ]
        ) == [{"subscriberPackageVersionId": "04t000000000000"}]

    def test_unpackaged_pre_dependencies__none(self, task):
        shutil.rmtree(str(pathlib.Path(task.project_config.repo_root, "unpackaged")))

        assert task._get_unpackaged_pre_dependencies([]) == []

    def test_unpackaged_pre_dependencies__no_unlocked_packages(self, task):
        task.options["create_unlocked_dependency_packages"] = False

        assert task._get_unpackaged_pre_dependencies([]) == []

    @responses.activate
    def test_poll_action__error(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [{"Id": "08c000000000002AAA", "Status": "Error"}],
            },
        )
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Message": "message"}]},
        )

        task.request_id = "08c000000000002AAA"
        with pytest.raises(PackageUploadFailure) as err:
            task._poll_action()
        assert "message" in str(err)

    @responses.activate
    def test_poll_action__other(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [{"Id": "08c000000000002AAA", "Status": "InProgress"}],
            },
        )

        task.request_id = "08c000000000002AAA"
        task._poll_action()

    @responses.activate
    def test_get_base_version_number__fallback(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        version = task._get_base_version_number(None, "0Ho6g000000fy4ZCAQ")
        assert version.format() == "0.0.0.0"

    @responses.activate
    def test_get_base_version_number__from_github(self, task):
        task.project_config.get_latest_version = mock.Mock(return_value="1.0.0.1")

        version = task._get_base_version_number(
            "latest_github_release", "0Ho6g000000fy4ZCAQ"
        )
        assert version.format() == "1.0.0.1"

    @responses.activate
    def test_get_base_version_number__from_github_1gp(self, task):
        task.project_config.get_latest_version = mock.Mock(return_value="1.0.0")

        version = task._get_base_version_number(
            "latest_github_release", "0Ho6g000000fy4ZCAQ"
        )
        assert version.format() == "1.0.0.0"

    @responses.activate
    def test_get_base_version_number__from_github_1gp_2_figures(self, task):
        task.project_config.get_latest_version = mock.Mock(return_value="1.0")

        version = task._get_base_version_number(
            "latest_github_release", "0Ho6g000000fy4ZCAQ"
        )
        assert version.format() == "1.0.0.0"

    @responses.activate
    def test_get_base_version_number__from_github_1gp_beta(self, task):
        # This shouldn't happen unless the project is misconfigured,
        # but we'll ensure we handle it gracefully.
        task.project_config.get_latest_version = mock.Mock(return_value="1.0 (Beta 2)")

        version = task._get_base_version_number(
            "latest_github_release", "0Ho6g000000fy4ZCAQ"
        )
        assert version.format() == "1.0.0.2"

    @responses.activate
    def test_get_base_version_number__from_github__no_release(self, task):
        task.project_config.get_latest_version = mock.Mock(side_effect=GithubException)

        version = task._get_base_version_number(
            "latest_github_release", "0Ho6g000000fy4ZCAQ"
        )
        assert version.format() == "0.0.0.0"

    @responses.activate
    def test_get_base_version_number__explicit(self, task):
        version = task._get_base_version_number("1.0.0.1", "0Ho6g000000fy4ZCAQ")
        assert version.format() == "1.0.0.1"

    @responses.activate
    def test_increment_major_version__no_version_base_specified(self, task):
        """Test incrementing version from 0.0.0.12 -> 1.0.0.0"""
        responses.add(  # query to find base version
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "Id": "04t000000000002AAA",
                        "MajorVersion": 0,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "BuildNumber": 12,
                        "IsReleased": False,
                    }
                ],
            },
        )
        version_base = None
        version = task._get_base_version_number(version_base, "a package 2 Id")
        next_version = version.increment(VersionTypeEnum.major)
        assert next_version.format() == "1.0.0.NEXT"

    @responses.activate
    @mock.patch("cumulusci.tasks.create_package_version.get_version_id_from_tag")
    @mock.patch("cumulusci.tasks.create_package_version.get_latest_tag")
    @mock.patch("cumulusci.tasks.create_package_version.get_repo_from_config")
    def test_resolve_ancestor_id__latest_github_release(
        self, get_repo_from_config, get_latest_tag, get_version_id_from_tag, task
    ):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Id": "05i000000000000"}]},
        )

        project_config = mock.Mock()
        task.project_config = project_config

        # Mock the repo object
        mock_repo = mock.Mock()
        get_repo_from_config.return_value = mock_repo
        get_latest_tag.return_value = "v1.0.0"
        get_version_id_from_tag.return_value = "04t000000000111"

        actual_id = task._resolve_ancestor_id("latest_github_release")
        assert actual_id == "05i000000000000"

    @responses.activate
    def test_resolve_ancestor_id__no_ancestor_specified(self, task):
        project_config = mock.Mock()
        project_config.get_latest_tag.side_effect = GithubException
        task.project_config = project_config

        assert task._resolve_ancestor_id() == ""

    @responses.activate
    @mock.patch("cumulusci.tasks.create_package_version.get_version_id_from_tag")
    def test_resolve_ancestor_id__ancestor_explicitly_specified(
        self, get_version_id_from_tag, task
    ):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 1, "records": [{"Id": "05i000000000000"}]},
        )

        project_config = mock.Mock()
        task.project_config = project_config

        get_version_id_from_tag.return_value = "04t000000000111"

        actual_id = task._resolve_ancestor_id("04t000000000000")
        assert actual_id == "05i000000000000"

    @responses.activate
    @mock.patch("cumulusci.tasks.create_package_version.get_latest_tag")
    @mock.patch("cumulusci.tasks.create_package_version.get_repo_from_config")
    def test_resolve_ancestor_id__no_release_found(
        self, get_repo_from_config, get_latest_tag, task
    ):
        project_config = mock.Mock()
        project_config.get_latest_tag.side_effect = GithubException
        task.project_config = project_config

        # Mock the repo object
        mock_repo = mock.Mock()
        get_repo_from_config.return_value = mock_repo
        get_latest_tag.side_effect = VcsException("No release found")

        assert task._resolve_ancestor_id("latest_github_release") == ""

    def test_resolve_ancestor_id__unlocked_package(self, task):
        task.package_config = PackageConfig(
            package_name="test_package",
            package_type="Unlocked",
            org_dependent=False,
            post_install_script=None,
            uninstall_script=None,
            namespace="test",
            version_name="Release",
            version_base=None,
            version_type="patch",
        )
        with pytest.raises(
            CumulusCIUsageError,
            match="Cannot specify an ancestor for Unlocked packages.",
        ):
            task._resolve_ancestor_id("04t000000000000")

    def test_resolve_ancestor_id__invalid_option_value(self, task):
        with pytest.raises(
            TaskOptionsError,
            match=re.escape("Unrecognized value for ancestor_id: 001001001001001"),
        ):
            task._resolve_ancestor_id("001001001001001")

    def test_prepare_cci_dependencies(self, task):
        assert task._prepare_cci_dependencies("") == []
        assert task._prepare_cci_dependencies(None) == []
        assert task._prepare_cci_dependencies(
            {"ids": [{"subscriberPackageVersionId": "04t000000000000"}]}
        ) == [{"version_id": "04t000000000000"}]


class TestPackageConfigNewFeatures:
    """Tests for new PackageConfig validation features."""

    def test_validate_apex_test_access(self):
        with pytest.raises(ValidationError, match="Only managed packages"):
            PackageConfig(
                package_type=PackageTypeEnum.unlocked,
                apex_test_access={"permission_set_names": ["TestPermSet"]},
            )  # type: ignore

    def test_validate_package_metadata_access(self):
        with pytest.raises(ValidationError, match="Only managed packages"):
            PackageConfig(
                package_type=PackageTypeEnum.unlocked,
                package_metadata_access={"permission_set_names": ["TestPermSet"]},
            )  # type: ignore


class TestCreatePackageVersionNewFeatures:
    """Tests for new CreatePackageVersion functionality."""

    devhub_base_url = (
        f"https://devhub.my.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}"
    )

    @responses.activate
    def test_version_number_option(self, get_task, devhub_config):
        """Test that version_number option is used instead of version_base/version_type"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
                "version_number": "2.5.3.10",
            }
        )

        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )
        responses.add(
            "POST",
            f"{self.devhub_base_url}/tooling/sobjects/Package2/",
            json={"id": "0Ho6g000000fy4ZCAQ"},
        )
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        package_id = task._get_or_create_package(task.package_config)
        builder = BasePackageZipBuilder()

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as _, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                # Verify version_number was used instead of version_base/version_type
                assert descriptor["versionNumber"] == "2.5.3.10"
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                package_id, task.package_config, builder, skip_validation=True
            )
            # Verify version_number was parsed correctly
            assert task.options["version_number"] is not None
            assert task.options["version_number"].format() == "2.5.3.10"

    @responses.activate
    def test_dependencies_option(self, get_task, devhub_config):
        """Test that explicit dependencies option is used"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
                "dependencies": "04t000000000001,04t000000000002",
            }
        )

        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )
        responses.add(
            "POST",
            f"{self.devhub_base_url}/tooling/sobjects/Package2/",
            json={"id": "0Ho6g000000fy4ZCAQ"},
        )
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        package_id = task._get_or_create_package(task.package_config)
        builder = BasePackageZipBuilder()

        with mock.patch.object(
            task, "_get_dependencies"
        ) as mock_get_deps, mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                # Verify dependencies were set correctly
                assert "dependencies" in descriptor
                assert descriptor["dependencies"] == [
                    {"subscriberPackageVersionId": "04t000000000001"},
                    {"subscriberPackageVersionId": "04t000000000002"},
                ]
                # _get_dependencies should not be called when dependencies option is set
                mock_get_deps.assert_not_called()
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                package_id,
                task.package_config,
                builder,
                skip_validation=False,
            )
            # Verify dependencies were parsed correctly
            assert task.options["dependencies"] == [
                {"subscriberPackageVersionId": "04t000000000001"},
                {"subscriberPackageVersionId": "04t000000000002"},
            ]

    @responses.activate
    def test_apex_test_access_string_format(self, get_task, devhub_config):
        """Test apex_test_access with string format for permission sets"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.apex_test_access = {
            "permission_set_names": "PermSet1, PermSet2",
            "permission_set_license_names": "License1, License2",
        }

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                assert "permissionSetNames" in descriptor
                assert descriptor["permissionSetNames"] == ["PermSet1", "PermSet2"]
                assert "permissionSetLicenseDeveloperNames" in descriptor
                assert descriptor["permissionSetLicenseDeveloperNames"] == [
                    "License1",
                    "License2",
                ]
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )

    @responses.activate
    def test_apex_test_access_list_format(self, get_task, devhub_config):
        """Test apex_test_access with list format for permission sets"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.apex_test_access = {
            "permission_set_names": ["PermSet1", "PermSet2"],
            "permission_set_license_names": ["License1", "License2"],
        }

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                assert "permissionSetNames" in descriptor
                assert descriptor["permissionSetNames"] == ["PermSet1", "PermSet2"]
                assert "permissionSetLicenseDeveloperNames" in descriptor
                assert descriptor["permissionSetLicenseDeveloperNames"] == [
                    "License1",
                    "License2",
                ]
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )

    @responses.activate
    def test_package_metadata_access_string_format(self, get_task, devhub_config):
        """Test package_metadata_access with string format"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.package_metadata_access = {
            "permission_set_names": "MetaPermSet1, MetaPermSet2",
            "permission_set_license_names": "MetaLicense1, MetaLicense2",
        }

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                assert "packageMetadataPermissionSetNames" in descriptor
                assert descriptor["packageMetadataPermissionSetNames"] == [
                    "MetaPermSet1",
                    "MetaPermSet2",
                ]
                assert "packageMetadataPermissionSetLicenseNames" in descriptor
                assert descriptor["packageMetadataPermissionSetLicenseNames"] == [
                    "MetaLicense1",
                    "MetaLicense2",
                ]
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )

    @responses.activate
    def test_package_metadata_access_list_format(self, get_task, devhub_config):
        """Test package_metadata_access with list format"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.package_metadata_access = {
            "permission_set_names": ["MetaPermSet1", "MetaPermSet2"],
            "permission_set_license_names": ["MetaLicense1", "MetaLicense2"],
        }

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                assert "packageMetadataPermissionSetNames" in descriptor
                assert descriptor["packageMetadataPermissionSetNames"] == [
                    "MetaPermSet1",
                    "MetaPermSet2",
                ]
                assert "packageMetadataPermissionSetLicenseNames" in descriptor
                assert descriptor["packageMetadataPermissionSetLicenseNames"] == [
                    "MetaLicense1",
                    "MetaLicense2",
                ]
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )

    @responses.activate
    @mock.patch("cumulusci.tasks.create_package_version.consolidate_metadata")
    @mock.patch("cumulusci.tasks.create_package_version.convert_sfdx_source")
    @mock.patch("cumulusci.tasks.create_package_version.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.tasks.create_package_version.clean_temp_directory")
    def test_get_unpackaged_metadata_path_string(
        self,
        mock_clean_temp,
        mock_zip_builder,
        mock_convert_sfdx,
        mock_consolidate,
        task,
    ):
        """Test _get_unpackaged_metadata_path with string path"""
        import tempfile

        temp_path = tempfile.mkdtemp()
        mock_consolidate.return_value = temp_path, 1
        mock_convert_sfdx.return_value.__enter__.return_value = temp_path

        mock_builder_instance = mock.Mock()
        mock_builder_instance.as_bytes.return_value = b"testzipbytes"
        mock_zip_builder.return_value = mock_builder_instance

        version_bytes = io.BytesIO()
        version_info = zipfile.ZipFile(version_bytes, "w", zipfile.ZIP_DEFLATED)

        result = task._get_unpackaged_metadata_path("unpackaged/pre", version_info)

        assert result == version_info
        mock_consolidate.assert_called_once_with(
            "unpackaged/pre", task.project_config.repo_root, logger=task.logger
        )
        mock_convert_sfdx.assert_called_once()
        mock_zip_builder.assert_called_once()
        mock_builder_instance.as_bytes.assert_called_once()
        mock_clean_temp.assert_called_once_with(temp_path)
        assert "unpackaged-metadata-package.zip" in [
            name for name in version_info.namelist()
        ]

        version_info.close()

    @responses.activate
    @mock.patch("cumulusci.tasks.create_package_version.consolidate_metadata")
    @mock.patch("cumulusci.tasks.create_package_version.convert_sfdx_source")
    @mock.patch("cumulusci.tasks.create_package_version.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.tasks.create_package_version.clean_temp_directory")
    def test_get_unpackaged_metadata_path_list(
        self,
        mock_clean_temp,
        mock_zip_builder,
        mock_convert_sfdx,
        mock_consolidate,
        task,
    ):
        """Test _get_unpackaged_metadata_path with list of paths"""
        import tempfile

        temp_path = tempfile.mkdtemp()
        mock_consolidate.return_value = temp_path, 1
        mock_convert_sfdx.return_value.__enter__.return_value = temp_path

        mock_builder_instance = mock.Mock()
        mock_builder_instance.as_bytes.return_value = b"testzipbytes"
        mock_zip_builder.return_value = mock_builder_instance

        version_bytes = io.BytesIO()
        version_info = zipfile.ZipFile(version_bytes, "w", zipfile.ZIP_DEFLATED)

        metadata_paths = ["unpackaged/pre", "unpackaged/post"]
        result = task._get_unpackaged_metadata_path(metadata_paths, version_info)

        assert result == version_info
        mock_consolidate.assert_called_once_with(
            metadata_paths, task.project_config.repo_root, logger=task.logger
        )
        version_info.close()

    @responses.activate
    @mock.patch("cumulusci.tasks.create_package_version.consolidate_metadata")
    @mock.patch("cumulusci.tasks.create_package_version.convert_sfdx_source")
    @mock.patch("cumulusci.tasks.create_package_version.MetadataPackageZipBuilder")
    @mock.patch("cumulusci.tasks.create_package_version.clean_temp_directory")
    def test_get_unpackaged_metadata_path_dict(
        self,
        mock_clean_temp,
        mock_zip_builder,
        mock_convert_sfdx,
        mock_consolidate,
        task,
    ):
        """Test _get_unpackaged_metadata_path with dict format"""
        import tempfile

        temp_path = tempfile.mkdtemp()
        mock_consolidate.return_value = temp_path, 1
        mock_convert_sfdx.return_value.__enter__.return_value = temp_path

        mock_builder_instance = mock.Mock()
        mock_builder_instance.as_bytes.return_value = b"testzipbytes"
        mock_zip_builder.return_value = mock_builder_instance

        version_bytes = io.BytesIO()
        version_info = zipfile.ZipFile(version_bytes, "w", zipfile.ZIP_DEFLATED)

        metadata_paths = {"unpackaged/pre": "*.*", "unpackaged/post": "test.xml"}
        result = task._get_unpackaged_metadata_path(metadata_paths, version_info)

        assert result == version_info
        mock_consolidate.assert_called_once_with(
            metadata_paths, task.project_config.repo_root, logger=task.logger
        )
        version_info.close()

    @responses.activate
    def test_create_version_request_with_unpackaged_metadata(
        self, get_task, devhub_config
    ):
        """Test _create_version_request includes unpackaged metadata when configured"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.unpackaged_metadata_path = "unpackaged/pre"

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_unpackaged_metadata_path"
        ) as mock_unpackaged, mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            version_bytes = io.BytesIO()
            version_info = zipfile.ZipFile(version_bytes, "w", zipfile.ZIP_DEFLATED)
            mock_unpackaged.return_value = version_info

            mock_tooling_obj = mock.Mock()
            mock_tooling_obj.create.return_value = {"id": "08c000000000002AAA"}
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )

            mock_unpackaged.assert_called_once_with("unpackaged/pre", mock.ANY)
            version_info.close()

    @responses.activate
    def test_apex_test_access_partial_config(self, get_task, devhub_config):
        """Test apex_test_access with only permission_set_names"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.apex_test_access = {
            "permission_set_names": ["PermSet1"],
        }

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                assert "permissionSetNames" in descriptor
                assert "permissionSetLicenseDeveloperNames" not in descriptor
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )

    @responses.activate
    def test_package_metadata_access_partial_config(self, get_task, devhub_config):
        """Test package_metadata_access with only permission_set_license_names"""
        task = get_task(
            {
                "package_type": "Managed",
                "package_name": "Test Package",
            }
        )
        task.package_config.package_metadata_access = {
            "permission_set_license_names": ["MetaLicense1"],
        }

        builder = BasePackageZipBuilder()

        # Mock the query for existing Package2VersionCreateRequest
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": []},
        )

        with mock.patch.object(
            task, "_get_base_version_number"
        ) as mock_version, mock.patch.object(
            builder, "as_hash", return_value="testhash"
        ), mock.patch.object(
            builder, "as_bytes", return_value=b"testbytes"
        ), mock.patch.object(
            task, "_get_tooling_object"
        ) as mock_tooling:
            mock_version.return_value.increment.return_value.format.return_value = (
                "1.0.0.1"
            )
            mock_tooling_obj = mock.Mock()

            def capture_create(request):
                import base64

                version_info_b64 = request["VersionInfo"]
                version_info_bytes = base64.b64decode(version_info_b64)
                version_info_zip = zipfile.ZipFile(io.BytesIO(version_info_bytes), "r")
                descriptor_json = version_info_zip.read("package2-descriptor.json")
                descriptor = json.loads(descriptor_json.decode("utf-8"))
                assert "packageMetadataPermissionSetLicenseNames" in descriptor
                assert "packageMetadataPermissionSetNames" not in descriptor
                return {"id": "08c000000000002AAA"}

            mock_tooling_obj.create.side_effect = capture_create
            mock_tooling.return_value = mock_tooling_obj

            task._create_version_request(
                "0Ho6g000000fy4ZCAQ",
                task.package_config,
                builder,
                skip_validation=True,
            )



""" FlowRunner contains the logic for actually running a flow.

Flows are an integral part of CCI, they actually *do the thing*. We've been getting
along quite nicely with BaseFlow, which turns a flow definition into a callable
object that runs the flow in one fell swoop. We named it BaseFlow thinking that,
like tasks, specific flows might subclass it to extend behavior. In practice,
unlike BaseTask, subclasses ended up representing variations in how the flow
should actually be executed. We added callback hooks like pre_task and post_task
for host systems embedding cci, like web apps, to inspect the flow in progress.

BaseFlow suited us well.

FlowRunner is a v2 API for flows in CCI. There are two objects of interest:

- FlowCoordinator: takes a flow_config & runtime options to create a set of StepSpecs
  - Meant to replace the public API of BaseFlow, including override hooks.
  - Precomputes a flat list of steps, instead of running Flow recursively.
- TaskRunner: encapsulates the actual task running, result providing logic.

Upon initialization, FlowRunner:

- Creates a logger
- Validates that there are no cycles in the given flow_config
- Validates that the flow_config is using new-style-steps
- Collects a list of StepSpec objects that define what the flow will do.

Upon running the flow, FlowRunner:

- Refreshes the org credentials
- Runs each StepSpec in order
- * Logs the task or skip
- * Updates any ^^ task option values with return_values references
- * Creates a TaskRunner to run the task and get the result
- * Re-raise any fatal exceptions from the task, if not ignore_failure.
- * collects StepResults into the flow.

TaskRunner:

- Imports the actual task module.
- Constructs an instance of the BaseTask subclass.
- Runs/calls the task instance.
- Returns results or exception into an immutable StepResult

Option values/overrides can be passed in at a number of levels, in increasing order of priority:

- Task default (i.e. `.tasks__TASKNAME__options`)
- Flow definition task options (i.e. `.flows__FLOWNAME__steps__STEPNUM__options`)
- Flow definition subflow options (i.e. `.flows__FLOWNAME__steps__STEPNUM__options__TASKNAME`)
    see `dev_org_namespaced` for an example
- Flow runtime (i.e. on the commandline)

"""

import copy
import logging
import os
from collections import defaultdict
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
)

from jinja2.sandbox import ImmutableSandboxedEnvironment

from cumulusci.core.config import FlowConfig, TaskConfig
from cumulusci.core.config.org_config import OrgConfig
from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.exceptions import (
    FlowConfigError,
    FlowInfiniteLoopError,
    TaskImportError,
)
from cumulusci.utils.version_strings import LooseVersion

if TYPE_CHECKING:
    from cumulusci.core.tasks import BaseTask


RETURN_VALUE_OPTION_PREFIX = "^^"

jinja2_env = ImmutableSandboxedEnvironment()


class StepVersion(LooseVersion):
    """Like LooseVersion, but converts "/" into -1 to support comparisons"""

    def parse(self, vstring: str):
        super().parse(vstring)
        self.version = tuple(-1 if x == "/" else x for x in self.version)


class StepSpec:
    """simple namespace to describe what the flowrunner should do each step"""

    __slots__ = (
        "step_num",
        "task_name",
        "task_config",
        "task_class",
        "project_config",
        "allow_failure",
        "path",
        "skip",
        "when",
    )

    step_num: StepVersion
    task_name: str
    task_config: dict
    task_class: Optional[
        Type["BaseTask"]
    ]  # None means this step was skipped by setting task: None
    project_config: BaseProjectConfig
    allow_failure: bool
    path: str
    skip: bool
    when: Optional[str]

    def __init__(
        self,
        step_num: StepVersion,
        task_name: str,
        task_config: dict,
        task_class: Optional[Type["BaseTask"]],
        project_config: BaseProjectConfig,
        allow_failure: bool = False,
        from_flow: Optional[str] = None,
        skip: bool = False,
        when: Optional[str] = None,
    ):
        self.step_num = step_num
        self.task_name = task_name
        self.task_config = task_config
        self.task_class = task_class
        self.project_config = project_config
        self.allow_failure = allow_failure
        self.skip = skip
        self.when = when

        # Store the dotted path to this step.
        # This is not guaranteed to be unique, because multiple steps
        # in the same flow can reference the same task name with different options.
        # It's here to support the ^^flow_name.task_name.attr_name syntax
        # for referencing previous task return values in options.
        if from_flow:
            self.path = ".".join([from_flow, task_name])
        else:
            self.path = task_name

    def __repr__(self):
        skipstr = ""
        if self.skip:
            skipstr = "!SKIP! "
        return (
            f"<{skipstr}StepSpec {self.step_num}:{self.task_name} {self.task_config}>"
        )


class FlowStepSpec(StepSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class StepResult(NamedTuple):
    step_num: StepVersion
    task_name: str
    path: str
    result: Any
    return_values: Any
    exception: Optional[Exception]


class FlowCallback:
    """A subclass of FlowCallback allows code running a flow
    to inject callback methods to run during the flow. Anything you
    would like the FlowCallback to have access to can be passed to the
    constructor. This is typically used to pass a Django model or model id
    when running a flow inside of a web app.

    Example subclass of FlowCallback:

        class CustomFlowCallback(FlowCallback):
            def __init__(self, model):
                self.model = model

            def post_task(self, step, result):
                # do something to record state on self.model

    Once a subclass is defined, you can instantiate it, and
    pass it as the value for the 'callbacks' keyword argument
    when instantiating a FlowCoordinator.

    Example running a flow with custom callbacks:

        custom_callbacks = CustomFlowCallbacks(model_instance)
        flow_coordinator = FlowCoordinator(
            project_config,
            flow_config,
            name=flow_name,
            options=options,
            callbacks=custom_callbacks,
        )
        flow_coordinator.run(org_config)


    """

    def pre_flow(self, coordinator: "FlowCoordinator"):
        """This is passed an instance of FlowCoordinator,
        that pertains to the flow which is about to run."""
        pass

    def post_flow(self, coordinator: "FlowCoordinator"):
        """This is passed an instance of FlowCoordinator,
        that pertains to the flow just finished running.
        This step executes whether or not the flow completed
        successfully."""
        pass

    def pre_task(self, step: StepSpec):
        """This is passed an instance StepSpec, that
        pertains to the task which is about to run."""
        pass

    def post_task(self, step: StepSpec, result: StepResult):
        """This method is called after a task has executed.

        :param step: Instance of StepSpec that relates to the task which executed
        :param result: Instance of the StepResult class that was run. Attributes of
        interest include, `result.result`, `result.return_values`, and `result.exception`
        """
        pass


class TaskRunner:
    """TaskRunner encapsulates the job of instantiating and running a task."""

    step: StepSpec
    org_config: Optional[OrgConfig]
    flow: Optional["FlowCoordinator"]

    def __init__(
        self,
        step: StepSpec,
        org_config: Optional[OrgConfig],
        flow: Optional["FlowCoordinator"] = None,
    ):
        self.step = step
        self.org_config = org_config
        self.flow = flow

    @classmethod
    def from_flow(cls, flow: "FlowCoordinator", step: StepSpec) -> "TaskRunner":
        return cls(step, flow.org_config, flow=flow)

    def run_step(self, **options) -> StepResult:
        """
        Run a step.

        :return: StepResult
        """

        # Resolve ^^task_name.return_value style option syntax
        task_config = self.step.task_config.copy()
        task_config["options"] = task_config.get("options", {}).copy()
        assert self.flow
        self.flow.resolve_return_value_options(task_config["options"])

        task_config["options"].update(options)

        assert self.step.task_class

        task = self.step.task_class(
            self.step.project_config,
            TaskConfig(task_config),
            org_config=self.org_config,
            name=self.step.task_name,
            stepnum=self.step.step_num,
            flow=self.flow,
        )
        self._log_options(task)
        exc = None
        try:
            task()
        except Exception as e:
            self.flow.logger.error(f"Exception in task {self.step.path}")
            exc = e
        return StepResult(
            self.step.step_num,
            self.step.task_name,
            self.step.path,
            task.result,
            task.return_values,
            exc,
        )

    def _log_options(self, task: "BaseTask"):
        if not task.task_options:
            task.logger.info("No task options present")
            return
        task.logger.info("Options:")
        for key, info in task.task_options.items():
            value = task.options.get(key)
            if value is not None:
                if type(value) is not list:
                    value = self._obfuscate_if_sensitive(value, info)
                    task.logger.info(f"  {key}: {value}")
                else:
                    task.logger.info(f"  {key}:")
                    for v in value:
                        v = self._obfuscate_if_sensitive(v, info)
                        task.logger.info(f"    - {v}")

    def _obfuscate_if_sensitive(self, value: str, info: dict) -> str:
        if info.get("sensitive"):
            value = 8 * "*"
        return value


class FlowCoordinator:
    org_config: Optional[OrgConfig]
    steps: List[StepSpec]
    callbacks: FlowCallback
    logger: logging.Logger
    skip: List[str]
    flow_config: FlowConfig
    runtime_options: dict
    name: Optional[str]
    results: List[StepResult]

    def __init__(
        self,
        project_config: BaseProjectConfig,
        flow_config: FlowConfig,
        name: Optional[str] = None,
        options: Optional[dict] = None,
        skip: Optional[List[str]] = None,
        callbacks: Optional[FlowCallback] = None,
    ):
        self.project_config = project_config
        self.flow_config = flow_config
        self.name = name
        self.org_config = None

        if not callbacks:
            callbacks = FlowCallback()
        self.callbacks = callbacks

        self.runtime_options = options or {}

        self.skip = skip or []
        self.results = []

        self.logger = self._init_logger()
        self.steps = self._init_steps()
        self._expression_cache = {}
        self._jinja2_context = None
        self._context_project_config = None
        self._context_org_config = None

    @classmethod
    def from_steps(
        cls,
        project_config: BaseProjectConfig,
        steps: List[StepSpec],
        name: Optional[str] = None,
        callbacks: Optional[FlowCallback] = None,
    ):
        instance = cls(
            project_config,
            flow_config=FlowConfig({"steps": {}}),
            name=name,
            callbacks=callbacks,
        )
        instance.steps = steps
        return instance

    def _rule(self, fill="=", length=60, new_line=False):
        self.logger.info(f"{fill * length}")
        if new_line:
            self.logger.info("")

    def get_summary(self, verbose=False):
        """Returns an output string that contains the description of the flow
        and its steps."""
        lines = []
        if "description" in self.flow_config.config:
            lines.append(f"Description: {self.flow_config.config['description']}")

        step_lines = self.get_flow_steps(verbose=verbose)
        if step_lines:
            lines.append("\nFlow Steps")
        lines.extend(step_lines)

        return "\n".join(lines)

    def get_flow_steps(
        self, for_docs: bool = False, verbose: bool = False
    ) -> List[str]:
        """Returns a list of flow steps (tasks and sub-flows) for the given flow.
        For docs, indicates whether or not we want to use the string for use in a code-block
        of an rst file. If True, will omit output of source information."""
        lines = []
        previous_parts = []
        previous_source = None
        for step in self.steps:
            if isinstance(step, FlowStepSpec):
                continue

            parts = step.path.split(".")
            steps = str(step.step_num).split("/")
            if len(parts) > len(steps):
                # Sub-step generated during freeze process; skip it
                continue
            task_name = parts.pop()

            i = -1
            new_source = (
                f" [from {step.project_config.source}]"
                if step.project_config.source is not previous_source
                else ""
            )
            options_info = ""
            for i, flow_name in enumerate(parts):
                if not any(":" in part for part in step.path.split(".")[i + 1 :]):
                    source = new_source
                else:
                    source = ""
                if len(previous_parts) < i + 1 or previous_parts[i] != flow_name:
                    if for_docs:
                        source = ""

                    lines.append(f"{'    ' * i}{steps[i]}) flow: {flow_name}{source}")
                    if source:
                        new_source = ""

            padding = "    " * (i + 1) + " " * len(str(steps[i + 1]))
            when = f"{padding}  when: {step.when}" if step.when is not None else ""

            if for_docs:
                new_source = ""

            if step.task_config.get("options"):
                if verbose:
                    options = step.task_config.get("options")
                    options_info = f"{padding}  options:"

                    for option, value in options.items():
                        options_info += f"\n{padding}      {option}: {value}"

            lines.append(
                f"{'    ' * (i + 1)}{steps[i + 1]}) task: {task_name}{new_source}"
            )

            if when:
                lines.append(when)

            if options_info:
                lines.append(options_info)

            previous_parts = parts
            previous_source = step.project_config.source

        return lines

    def run(self, org_config: OrgConfig):
        self.org_config = org_config
        line = f"Initializing flow: {self.__class__.__name__}"
        if self.name:
            line = f"{line} ({self.name})"
        self._rule()
        self.logger.info(line)
        self.logger.info(self.flow_config.description)
        self._rule(new_line=True)

        if self.org_config:
            self._init_org()
            self._rule(fill="-")
            self.logger.info("Organization:")
            self.logger.info(f"  Username: {org_config.username}")
            self.logger.info(f"    Org Id: {org_config.org_id}")
            self.logger.info(f"  Instance: {org_config.instance_name}")
            self._rule(fill="-", new_line=True)

        # Give pre_flow callback a chance to alter the steps
        # based on the state of the org before we display the steps.
        self.callbacks.pre_flow(self)

        self._rule(fill="-")
        self.logger.info("Steps:")
        for line in self.get_summary().splitlines():
            self.logger.info(line)
        self._rule(fill="-", new_line=True)

        self.logger.info("Starting execution")
        self._rule(new_line=True)

        try:
            # Pre-evaluate all flow conditions
            skipped_flows_set = set()
            for step in self.steps:
                if isinstance(step, FlowStepSpec):
                    if not self._evaluate_flow_step(step):
                        skipped_flows_set.add(step.path)

            # Main execution loop with optimized path checking
            for step in self.steps:
                if isinstance(step, FlowStepSpec):
                    self.logger.info(
                        f"Skipping Flow {step.task_name} (skipped unless {step.when})"
                    )
                    continue

                if self._is_task_in_skipped_flow(step.path, skipped_flows_set):
                    self.logger.info(
                        f"Skipping Task {step.task_name} in flow {step.path} (parent flow is skipped)"
                    )
                    continue

                self._run_step(step)
            flow_name = f"'{self.name}' " if self.name else ""
            org_name = f"on org {org_config.name} " if org_config else ""
            self.logger.info(f"Completed flow {flow_name}{org_name}successfully!")
        finally:
            self.callbacks.post_flow(self)

    def _get_jinja2_context(self, project_config, org_config):
        """Get or create jinja2 context, reusing when possible."""
        if (
            self._jinja2_context is None
            or self._context_project_config != project_config
            or self._context_org_config != org_config
        ):

            self._jinja2_context = {
                "project_config": project_config,
                "org_config": org_config,
                "env": dict(os.environ),
            }
            self._context_project_config = project_config
            self._context_org_config = org_config

        return self._jinja2_context

    def _evaluate_flow_step(self, step: StepSpec) -> bool:
        if not step.when:
            return True

        # Check cache first
        if step.when in self._expression_cache:
            expr = self._expression_cache[step.when]
        else:
            expr = jinja2_env.compile_expression(step.when)
            self._expression_cache[step.when] = expr

        jinja2_context = self._get_jinja2_context(step.project_config, self.org_config)

        return expr(**jinja2_context)

    def _is_task_in_skipped_flow(self, task_path: str, skipped_flows_set: set) -> bool:
        """Check if task belongs to any skipped flow using O(1) set lookup."""
        for skipped_path in skipped_flows_set:
            if task_path.startswith(skipped_path + "."):
                return True
        return False

    def _run_step(self, step: StepSpec):
        if step.skip:
            self._rule(fill="*")
            self.logger.info(f"Skipping task: {step.task_name}")
            self._rule(fill="*", new_line=True)
            return

        if step.when:
            value = self._evaluate_flow_step(step)
            if not value:
                self.logger.info(
                    f"Skipping task {step.task_name} (skipped unless {step.when})"
                )
                return

        self._rule(fill="-")
        self.logger.info(f"Running task: {step.task_name}")
        self._rule(fill="-", new_line=True)

        self.callbacks.pre_task(step)
        result = TaskRunner.from_flow(self, step).run_step()
        self.callbacks.post_task(step, result)

        self.results.append(
            result
        )  # add even a failed result to the result set for the post flow

        if result.exception and not step.allow_failure:
            raise result.exception  # PY3: raise an exception type we control *from* this exception instead?

    def _init_logger(self) -> logging.Logger:
        """
        Returns a logging.Logger-like object to use for the duration of the flow. Tasks will receive this logger
        and getChild(class_name) to get a child logger.

        :return: logging.Logger
        """
        return logging.getLogger("cumulusci.flows").getChild(self.__class__.__name__)

    def _init_steps(self) -> List[StepSpec]:
        """
        Given the flow config and everything else, create a list of steps to run, sorted by step number.

        :return: List[StepSpec]
        """
        self._check_old_yaml_format()
        self._check_infinite_flows(self.flow_config)

        steps = []

        for number, step_config in self.flow_config.steps.items():
            specs = self._visit_step(number, step_config, self.project_config)
            steps.extend(specs)

        return sorted(steps, key=attrgetter("step_num"))

    def _visit_step(
        self,
        number: Union[str, int],
        step_config: dict,
        project_config: BaseProjectConfig,
        visited_steps: Optional[List[StepSpec]] = None,
        parent_options: Optional[dict] = None,
        parent_ui_options: Optional[dict] = None,
        from_flow: Optional[str] = None,
    ) -> List[StepSpec]:
        """
        for each step (as defined in the flow YAML), _visit_step is called with only
        the first two parameters. this takes care of validating the step, collating the
        option overrides, and if it is a task, creating a StepSpec for it.

        If it is a flow, we recursively call _visit_step with the rest of the parameters of context.

        :param number: StepVersion representation of the current step number
        :param step_config: the current step's config (dict from YAML)
        :param visited_steps: used when called recursively for nested steps, becomes the return value
        :param parent_options: used when called recursively for nested steps, options from parent flow
        :param parent_ui_options: used when called recursively for nested steps, UI options from parent flow
        :param from_flow: used when called recursively for nested steps, name of parent flow
        :return: List[StepSpec] a list of all resolved steps including/under the one passed in
        """
        step_number = StepVersion(str(number))

        if visited_steps is None:
            visited_steps = []
        if parent_options is None:
            parent_options = {}
        if parent_ui_options is None:
            parent_ui_options = {}

        # This should never happen because of cleanup
        # in core/utils/cleanup_old_flow_step_replace_syntax()
        assert step_config.keys() != {"task", "flow"}

        # Skips
        # - either in YAML (with the None string)
        # - or by providing a skip list to the FlowRunner at initialization.
        if (
            ("flow" in step_config and step_config["flow"] == "None")
            or ("task" in step_config and step_config["task"] == "None")
            or ("task" in step_config and step_config["task"] in self.skip)
        ):
            visited_steps.append(
                StepSpec(
                    step_num=step_number,
                    task_name=step_config.get("task", step_config.get("flow")),
                    task_config=step_config.get("options", {}),
                    task_class=None,
                    project_config=project_config,
                    from_flow=from_flow,
                    skip=True,  # someday we could use different vals for why skipped
                )
            )
            return visited_steps

        if "task" in step_config:
            name = step_config["task"]

            # get the base task_config from the project config, as a dict for easier manipulation.
            # will raise if the task doesn't exist / is invalid
            task_config = project_config.get_task(name)
            task_config_dict: dict = copy.deepcopy(task_config.config)
            if "options" not in task_config_dict:
                task_config_dict["options"] = {}

            # merge the options together, from task_config all the way down through parent_options
            # parent_options should have higher priority than step_config options
            step_overrides = copy.deepcopy(step_config.get("options", {}))
            parent_task_options = parent_options.get(name, {})
            step_overrides.update(parent_task_options)
            task_config_dict["options"].update(step_overrides)

            # merge UI options from task config and parent flow
            if "ui_options" not in task_config_dict:
                task_config_dict["ui_options"] = {}
            # parent_ui_options should have higher priority than step_config ui_options
            step_ui_overrides = copy.deepcopy(step_config.get("ui_options", {}))
            step_ui_overrides.update(parent_ui_options.get(name, {}))
            task_config_dict["ui_options"].update(step_ui_overrides)

            # merge checks from task config and flow step
            if "checks" not in task_config_dict:
                task_config_dict["checks"] = []
            task_config_dict["checks"].extend(step_config.get("checks", []))

            # merge runtime options
            if name in self.runtime_options:
                task_config_dict["options"].update(self.runtime_options[name])

            # get implementation class. raise/fail if it doesn't exist, because why continue
            try:
                task_class = task_config.get_class()
            except (ImportError, AttributeError, TaskImportError) as e:
                raise FlowConfigError(f"Task named {name} has bad classpath, {e}")

            visited_steps.append(
                StepSpec(
                    step_num=step_number,
                    task_name=name,
                    task_config=task_config_dict,
                    task_class=task_class,
                    project_config=task_config.project_config,
                    allow_failure=step_config.get("ignore_failure", False),
                    from_flow=from_flow,
                    when=step_config.get("when"),
                )
            )
            return visited_steps

        if "flow" in step_config:
            name = step_config["flow"]
            if from_flow:
                path = ".".join([from_flow, name])
            else:
                path = name
            step_options = step_config.get("options", {})
            parent_task_options = parent_options.get(name, {})
            step_options.update(parent_task_options)
            step_ui_options = step_config.get("ui_options", {})
            flow_config = project_config.get_flow(name)

            if step_config.get("when"):
                visited_steps.append(
                    FlowStepSpec(
                        task_config={},
                        step_num=step_number,
                        task_name=path,
                        task_class=None,
                        project_config=flow_config.project_config,
                        allow_failure=step_config.get("ignore_failure", False),
                        when=step_config.get("when"),
                    )
                )

            for sub_number, sub_stepconf in flow_config.steps.items():
                # append the flow number to the child number, since its a LooseVersion.
                # e.g. if we're in step 2.3 which references a flow with steps 1-5, it
                #   simply ends up as five steps: 2.3.1, 2.3.2, 2.3.3, 2.3.4, 2.3.5
                # TODO: how does this work with nested flowveride? what does defining step 2.3.2 later do?
                num = f"{number}/{sub_number}"
                self._visit_step(
                    number=num,
                    step_config=sub_stepconf,
                    project_config=flow_config.project_config,
                    visited_steps=visited_steps,
                    parent_options=step_options,
                    parent_ui_options=step_ui_options,
                    from_flow=path,
                )
        return visited_steps

    def _check_old_yaml_format(self):
        if self.flow_config.steps is None:
            if "tasks" in self.flow_config.config:
                raise FlowConfigError(
                    'Old flow syntax detected.  Please change from "tasks" to "steps" in the flow definition.'
                )
            else:
                raise FlowConfigError("No steps found in the flow definition")

    def _check_infinite_flows(self, flow_config, flow_stack=None):
        """
        Recursively loop through the flow_config and check if there are any cycles.

        :param flow_config: FlowConfig to traverse to find cycles/infinite loops
        :param flow_stack: list of flow signatures already visited
        :return: None
        """
        if not flow_stack:
            flow_stack = []
        project_config = flow_config.project_config

        for step in flow_config.steps.values():
            if "flow" in step:
                next_flow_name = step["flow"]
                if next_flow_name == "None":
                    continue

                next_flow_config = project_config.get_flow(next_flow_name)
                signature = (
                    hash(next_flow_config.project_config.source),
                    next_flow_config.name,
                )

                if signature in flow_stack:
                    raise FlowInfiniteLoopError(
                        f"Infinite flows detected with flow {next_flow_name}"
                    )
                flow_stack.append(signature)
                self._check_infinite_flows(next_flow_config, flow_stack)
                flow_stack.pop()

    def _init_org(self):
        """Test and refresh credentials to the org specified."""
        self.logger.info(
            f"Verifying and refreshing credentials for the specified org: {self.org_config.name}."
        )

        # attempt to refresh the token, this can throw...
        with self.org_config.save_if_changed():
            self.org_config.refresh_oauth_token(self.project_config.keychain)

    def resolve_return_value_options(self, options):
        """Handle dynamic option value lookups in the format ^^task_name.attr"""
        for key, value in options.items():
            if isinstance(value, str) and value.startswith(RETURN_VALUE_OPTION_PREFIX):
                path, name = value[len(RETURN_VALUE_OPTION_PREFIX) :].rsplit(".", 1)
                result = self._find_result_by_path(path)
                options[key] = result.return_values.get(name)

    def _find_result_by_path(self, path):
        for result in self.results:
            if result.path[-len(path) :] == path:
                return result
        raise NameError(f"Path not found: {path}")


class PreflightFlowCoordinator(FlowCoordinator):
    """Coordinates running preflight checks instead of the actual flow steps."""

    preflight_results: DefaultDict[Optional[str], List[dict]]
    _task_caches: Dict[BaseProjectConfig, "TaskCache"]

    def run(self, org_config: OrgConfig):
        self.org_config = org_config
        self.callbacks.pre_flow(self)

        self._init_org()
        self._rule(fill="-")
        self.logger.info("Organization:")
        self.logger.info(f"  Username: {org_config.username}")
        self.logger.info(f"    Org Id: {org_config.org_id}")
        self._rule(fill="-", new_line=True)

        self.logger.info("Running preflight checks...")
        self._rule(new_line=True)

        self.preflight_results = defaultdict(list)
        # Expose for test access
        self._task_caches = {self.project_config: TaskCache(self, self.project_config)}
        try:
            # flow-level checks
            jinja2_context = {
                "tasks": self._task_caches[self.project_config],
                "project_config": self.project_config,
                "org_config": self.org_config,
            }
            for check in self.flow_config.checks or []:
                result = self.evaluate_check(check, jinja2_context)
                if result:
                    self.preflight_results[None].append(result)

            # Step-level checks
            for step in self.steps:
                jinja2_context["project_config"] = step.project_config
                # Create a cache for this project config, if not present
                # and not equal to the root project config.
                # This accommodates cross-project preflight checks.
                if step.project_config not in self._task_caches:
                    self._task_caches[step.project_config] = TaskCache(
                        self, step.project_config
                    )
                jinja2_context["tasks"] = self._task_caches[step.project_config]
                for check in step.task_config.get("checks", []):
                    result = self.evaluate_check(check, jinja2_context)
                    if result:
                        self.preflight_results[str(step.step_num)].append(result)
        finally:
            self.callbacks.post_flow(self)

    def evaluate_check(
        self, check: dict, jinja2_context: Dict[str, Any]
    ) -> Optional[dict]:
        self.logger.info(f"Evaluating check: {check['when']}")
        expr = jinja2_env.compile_expression(check["when"])
        value = bool(expr(**jinja2_context))
        self.logger.info(f"Check result: {value}")
        if value:
            return {"status": check["action"], "message": check.get("message")}


class TaskCache:
    """Provides access to named tasks and caches their results.

    This is intended for use in a jinja2 expression context
    so that multiple expressions evaluated in the same context
    can avoid running a task more than once with the same options.
    """

    project_config: BaseProjectConfig
    results: Dict[Tuple[str, Tuple[Any]], Any]

    def __init__(self, flow: FlowCoordinator, project_config: BaseProjectConfig):
        self.flow = flow
        # Cross-project flows may include preflight checks
        # that depend on their local context.
        self.project_config = project_config
        self.results = {}

    def __getattr__(self, task_name: str):
        return CachedTaskRunner(self, task_name)


class CachedTaskRunner:
    """Runs a task and caches the result in a TaskCache"""

    cache: TaskCache
    task_name: str

    def __init__(self, cache: TaskCache, task_name: str):
        self.cache = cache
        self.task_name = task_name

    def __call__(self, **options: dict) -> Any:
        cache_key = (self.task_name, tuple(sorted(options.items())))
        if cache_key in self.cache.results:
            return self.cache.results[cache_key].return_values

        task_config = self.cache.project_config.tasks[self.task_name]
        task_class = TaskConfig(
            {**task_config, "project_config": self.cache.project_config}
        ).get_class()
        step = StepSpec(
            step_num=StepVersion("1"),
            task_name=self.task_name,
            task_config=task_config,
            task_class=task_class,
            project_config=self.cache.project_config,
        )
        self.cache.flow.callbacks.pre_task(step)
        result = TaskRunner(step, self.cache.flow.org_config, self.cache.flow).run_step(
            **options
        )
        self.cache.flow.callbacks.post_task(step, result)

        self.cache.results[cache_key] = result
        return result.return_values



from unittest import mock

import pytest
from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.exceptions import SalesforceException, TaskOptionsError
from cumulusci.tasks.salesforce.deprecate_package import DeprecatePackage

from .util import create_task


class TestDeprecatePackage:
    def test_init_task(self):
        """Test task initialization"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        assert task.tooling is not None

    def test_init_task_without_org_uses_devhub(self):
        """Test task initialization without org uses devhub"""
        from cumulusci.core.config import TaskConfig
        from cumulusci.tests.util import create_project_config

        project_config = create_project_config()
        task_config = TaskConfig({"options": {"package": "0Ho000000000000AAA"}})
        task = DeprecatePackage(
            project_config,
            task_config,
            org_config=None,
        )

        with mock.patch(
            "cumulusci.tasks.salesforce.deprecate_package.get_devhub_config"
        ) as mock_get_devhub:
            with mock.patch.object(task, "_init_api", return_value=mock.Mock()):
                with mock.patch.object(task, "_init_bulk", return_value=mock.Mock()):
                    mock_devhub_config = mock.Mock()
                    mock_get_devhub.return_value = mock_devhub_config
                    task._init_task()

                    mock_get_devhub.assert_called_once_with(project_config)
                    assert task.org_config == mock_devhub_config

    def test_init_task_with_org_uses_provided_org(self):
        """Test task initialization with org uses provided org"""
        from cumulusci.core.config import TaskConfig
        from cumulusci.core.config.org_config import OrgConfig
        from cumulusci.tests.util import DummyKeychain, create_project_config

        project_config = create_project_config()
        task_config = TaskConfig({"options": {"package": "0Ho000000000000AAA"}})
        provided_org = OrgConfig(
            {
                "instance_url": "https://test.salesforce.com",
                "id": "https://test.salesforce.com/ORG_ID/USER_ID",
                "access_token": "TOKEN",
                "org_id": "ORG_ID",
                "username": "test-cci@example.com",
            },
            "test",
            keychain=DummyKeychain(),
        )
        task = DeprecatePackage(
            project_config,
            task_config,
            org_config=provided_org,
        )

        with mock.patch(
            "cumulusci.tasks.salesforce.deprecate_package.get_devhub_config"
        ) as mock_get_devhub:
            task._init_task()

            # Should not call get_devhub_config when org is provided
            mock_get_devhub.assert_not_called()
            assert task.org_config == provided_org

    def test_get_package_by_id_success(self):
        """Test getting package by Id"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        task.tooling.query.return_value = {
            "size": 1,
            "records": [
                {
                    "Id": "0Ho000000000000AAA",
                    "Name": "TestPackage",
                    "ContainerOptions": "Unlocked",
                    "IsDeprecated": False,
                }
            ],
        }

        result = task._get_package("0Ho000000000000AAA")
        assert result["Id"] == "0Ho000000000000AAA"
        assert result["Name"] == "TestPackage"
        task.tooling.query.assert_called_once()
        assert "Id='0Ho000000000000AAA'" in task.tooling.query.call_args[0][0]

    def test_get_package_by_name_success(self):
        """Test getting package by Name"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "TestPackage",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        task.tooling.query.return_value = {
            "size": 1,
            "records": [
                {
                    "Id": "0Ho000000000000AAA",
                    "Name": "TestPackage",
                    "ContainerOptions": "Unlocked",
                    "IsDeprecated": False,
                }
            ],
        }

        result = task._get_package("TestPackage")
        assert result["Id"] == "0Ho000000000000AAA"
        assert result["Name"] == "TestPackage"
        assert "Name='TestPackage'" in task.tooling.query.call_args[0][0]
        assert "IsDeprecated = FALSE" in task.tooling.query.call_args[0][0]

    def test_get_package_not_found(self):
        """Test package not found"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        task.tooling.query.return_value = {"size": 0, "records": []}

        result = task._get_package("0Ho000000000000AAA")
        assert result is None

    def test_get_package_dev_hub_error(self):
        """Test dev hub error when querying package"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        error_response = SalesforceMalformedRequest(
            "url",
            400,
            "resource_name",
            [{"message": "Object type 'Package2' is not supported"}],
        )
        task.tooling.query.side_effect = error_response

        with pytest.raises(TaskOptionsError, match="Dev Hub"):
            task._get_package("0Ho000000000000AAA")

    def test_get_package_multiple_found(self):
        """Test multiple packages with same name"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "TestPackage",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        task.tooling.query.return_value = {
            "size": 2,
            "records": [
                {"Id": "0Ho000000000000AAA", "Name": "TestPackage"},
                {"Id": "0Ho000000000000BBB", "Name": "TestPackage"},
            ],
        }

        with pytest.raises(TaskOptionsError, match="Multiple packages"):
            task._get_package("TestPackage")

    def test_get_package_versions_success(self):
        """Test getting package versions"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        task.tooling.query.return_value = {
            "size": 2,
            "records": [
                {
                    "Id": "05i000000000001AAA",
                    "MajorVersion": 1,
                    "MinorVersion": 0,
                    "PatchVersion": 0,
                    "IsReleased": False,
                    "IsDeprecated": False,
                    "SubscriberPackageVersionId": "04t000000000001AAA",
                },
                {
                    "Id": "05i000000000002AAA",
                    "MajorVersion": 1,
                    "MinorVersion": 1,
                    "PatchVersion": 0,
                    "IsReleased": False,
                    "IsDeprecated": False,
                    "SubscriberPackageVersionId": "04t000000000002AAA",
                },
            ],
        }

        result = task._get_package_versions("0Ho000000000000AAA")
        assert len(result) == 2
        assert result[0]["Id"] == "05i000000000001AAA"
        assert "IsDeprecated = FALSE" in task.tooling.query.call_args[0][0]

    def test_get_package_versions_empty(self):
        """Test getting package versions when none exist"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()
        task.tooling.query.return_value = {"size": 0, "records": []}

        result = task._get_package_versions("0Ho000000000000AAA")
        assert result == []

    def test_check_can_deprecate_version_unlocked(self):
        """Test can deprecate unlocked package version"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()

        version = {
            "Id": "05i000000000001AAA",
            "IsReleased": True,
        }
        assert task._check_can_deprecate_version(version, "Unlocked") is True

    def test_check_can_deprecate_version_managed_not_released(self):
        """Test can deprecate managed package version that is not released"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()

        version = {
            "Id": "05i000000000001AAA",
            "IsReleased": False,
        }
        assert task._check_can_deprecate_version(version, "Managed") is True

    def test_check_can_deprecate_version_managed_released(self):
        """Test cannot deprecate managed package version that is released"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()

        version = {
            "Id": "05i000000000001AAA",
            "IsReleased": True,
        }
        assert task._check_can_deprecate_version(version, "Managed") is False

    def test_check_can_deprecate_package_unlocked(self):
        """Test can deprecate unlocked package"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()

        package_info = {"ContainerOptions": "Unlocked"}
        versions = [{"IsReleased": True}]
        assert task._check_can_deprecate_package(package_info, versions) is True

    def test_check_can_deprecate_package_managed_no_released_versions(self):
        """Test can deprecate managed package with no released versions"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()

        package_info = {"ContainerOptions": "Managed"}
        versions = [{"IsReleased": False}]
        assert task._check_can_deprecate_package(package_info, versions) is True

    def test_check_can_deprecate_package_managed_with_released_versions(self):
        """Test cannot deprecate managed package with released versions"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()

        package_info = {"ContainerOptions": "Managed"}
        versions = [{"IsReleased": True}]
        assert task._check_can_deprecate_package(package_info, versions) is False

    def test_deprecate_versions_success(self):
        """Test successful deprecation of versions"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        Package2Version = mock.Mock()
        Package2Version.update.return_value = {"success": True}
        task._get_tooling_object = mock.Mock(return_value=Package2Version)

        versions = [
            {
                "Id": "05i000000000001AAA",
                "MajorVersion": 1,
                "MinorVersion": 0,
                "PatchVersion": 0,
                "SubscriberPackageVersionId": "04t000000000001AAA",
            }
        ]

        task._deprecate_versions(versions)
        Package2Version.update.assert_called_once_with(
            "05i000000000001AAA", {"IsDeprecated": True}
        )

    def test_deprecate_versions_failure(self):
        """Test failure when deprecating versions"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        Package2Version = mock.Mock()
        Package2Version.update.side_effect = Exception("Update failed")
        task._get_tooling_object = mock.Mock(return_value=Package2Version)

        versions = [
            {
                "Id": "05i000000000001AAA",
                "MajorVersion": 1,
                "MinorVersion": 0,
                "PatchVersion": 0,
                "SubscriberPackageVersionId": "04t000000000001AAA",
            }
        ]

        with pytest.raises(SalesforceException, match="Failed to delete"):
            task._deprecate_versions(versions)

    def test_deprecate_package_success(self):
        """Test successful deprecation of package"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        Package2 = mock.Mock()
        Package2.update.return_value = {"success": True}
        task._get_tooling_object = mock.Mock(return_value=Package2)

        task._deprecate_package("0Ho000000000000AAA", "TestPackage")
        Package2.update.assert_called_once_with(
            "0Ho000000000000AAA", {"IsDeprecated": True}
        )

    def test_deprecate_package_failure(self):
        """Test failure when deprecating package"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        Package2 = mock.Mock()
        Package2.update.side_effect = Exception("Update failed")
        task._get_tooling_object = mock.Mock(return_value=Package2)

        with pytest.raises(SalesforceException, match="Failed to delete Package2"):
            task._deprecate_package("0Ho000000000000AAA", "TestPackage")

    @mock.patch("cumulusci.tasks.salesforce.deprecate_package.click.confirm")
    def test_run_task_with_versions_and_confirmation(self, mock_confirm):
        """Test full task execution with versions and user confirmation"""
        mock_confirm.return_value = True
        import click

        click.no_prompt = False

        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        # Mock queries
        task.tooling.query.side_effect = [
            # Get package
            {
                "size": 1,
                "records": [
                    {
                        "Id": "0Ho000000000000AAA",
                        "Name": "TestPackage",
                        "ContainerOptions": "Unlocked",
                        "IsDeprecated": False,
                    }
                ],
            },
            # Get versions
            {
                "size": 1,
                "records": [
                    {
                        "Id": "05i000000000001AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "IsReleased": False,
                        "IsDeprecated": False,
                        "SubscriberPackageVersionId": "04t000000000001AAA",
                    }
                ],
            },
        ]

        Package2Version = mock.Mock()
        Package2Version.update.return_value = {"success": True}
        Package2 = mock.Mock()
        Package2.update.return_value = {"success": True}
        task._get_tooling_object = mock.Mock(side_effect=[Package2Version, Package2])

        task._run_task()

        # Verify versions were deprecated
        Package2Version.update.assert_called_once_with(
            "05i000000000001AAA", {"IsDeprecated": True}
        )
        # Verify package was deprecated
        Package2.update.assert_called_once_with(
            "0Ho000000000000AAA", {"IsDeprecated": True}
        )
        # Verify confirmation was requested
        mock_confirm.assert_called_once()

    @mock.patch("cumulusci.tasks.salesforce.deprecate_package.click.confirm")
    def test_run_task_user_cancels(self, mock_confirm):
        """Test user cancellation"""
        mock_confirm.return_value = False
        import click

        click.no_prompt = False

        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        task.tooling.query.side_effect = [
            {
                "size": 1,
                "records": [
                    {
                        "Id": "0Ho000000000000AAA",
                        "Name": "TestPackage",
                        "ContainerOptions": "Unlocked",
                        "IsDeprecated": False,
                    }
                ],
            },
            {
                "size": 1,
                "records": [
                    {
                        "Id": "05i000000000001AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "IsReleased": False,
                        "IsDeprecated": False,
                        "SubscriberPackageVersionId": "04t000000000001AAA",
                    }
                ],
            },
        ]

        Package2Version = mock.Mock()
        task._get_tooling_object = mock.Mock(return_value=Package2Version)

        with pytest.raises(SalesforceException, match="canceled"):
            task._run_task()

        # Should not have deprecated anything
        Package2Version.update.assert_not_called()

    @mock.patch("cumulusci.tasks.salesforce.deprecate_package.click")
    def test_run_task_no_prompt(self, mock_click):
        """Test task execution without prompt"""
        mock_click.no_prompt = True
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
                "no_prompt": True,
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        task.tooling.query.side_effect = [
            {
                "size": 1,
                "records": [
                    {
                        "Id": "0Ho000000000000AAA",
                        "Name": "TestPackage",
                        "ContainerOptions": "Unlocked",
                        "IsDeprecated": False,
                    }
                ],
            },
            {"size": 0, "records": []},  # No versions
        ]

        Package2 = mock.Mock()
        Package2.update.return_value = {"success": True}
        task._get_tooling_object = mock.Mock(return_value=Package2)

        task._run_task()

        Package2.update.assert_called_once_with(
            "0Ho000000000000AAA", {"IsDeprecated": True}
        )

    def test_run_task_package_not_found(self):
        """Test task execution when package not found"""
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        task.tooling.query.side_effect = [
            {"size": 0, "records": []},  # Package not found
        ]

        with pytest.raises(TaskOptionsError, match="not found"):
            task._run_task()

    @mock.patch("cumulusci.tasks.salesforce.deprecate_package.click")
    def test_run_task_cannot_deprecate_released_managed_version(self, mock_click):
        """Test cannot deprecate released managed package version"""
        mock_click.no_prompt = True
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
                "no_prompt": True,
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        task.tooling.query.side_effect = [
            {
                "size": 1,
                "records": [
                    {
                        "Id": "0Ho000000000000AAA",
                        "Name": "TestPackage",
                        "ContainerOptions": "Managed",
                        "IsDeprecated": False,
                    }
                ],
            },
            {
                "size": 1,
                "records": [
                    {
                        "Id": "05i000000000001AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "IsReleased": True,
                        "IsDeprecated": False,
                        "SubscriberPackageVersionId": "04t000000000001AAA",
                    }
                ],
            },
        ]

        with pytest.raises(SalesforceException, match="Cannot delete released Managed"):
            task._run_task()

    @mock.patch("cumulusci.tasks.salesforce.deprecate_package.click")
    def test_run_task_cannot_deprecate_released_managed_package(self, mock_click):
        """Test cannot deprecate managed package with released versions"""
        mock_click.no_prompt = True
        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
                "no_prompt": True,
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        task.tooling.query.side_effect = [
            {
                "size": 1,
                "records": [
                    {
                        "Id": "0Ho000000000000AAA",
                        "Name": "TestPackage",
                        "ContainerOptions": "Managed",
                        "IsDeprecated": False,
                    }
                ],
            },
            {
                "size": 1,
                "records": [
                    {
                        "Id": "05i000000000001AAA",
                        "MajorVersion": 1,
                        "MinorVersion": 0,
                        "PatchVersion": 0,
                        "IsReleased": True,
                        "IsDeprecated": False,
                        "SubscriberPackageVersionId": "04t000000000001AAA",
                    }
                ],
            },
        ]

        Package2Version = mock.Mock()
        Package2Version.update.return_value = {"success": True}
        task._get_tooling_object = mock.Mock(return_value=Package2Version)

        # This should fail at the package level check
        with pytest.raises(SalesforceException, match="Cannot delete released Managed"):
            task._run_task()

    def test_run_task_no_versions_with_confirmation(self):
        """Test task execution with no versions but with confirmation"""
        import click

        click.no_prompt = False

        task = create_task(
            DeprecatePackage,
            {
                "package": "0Ho000000000000AAA",
            },
        )
        task._init_task()
        task.tooling = mock.Mock()

        with mock.patch(
            "cumulusci.tasks.salesforce.deprecate_package.click.confirm"
        ) as mock_confirm:
            mock_confirm.return_value = True

            task.tooling.query.side_effect = [
                {
                    "size": 1,
                    "records": [
                        {
                            "Id": "0Ho000000000000AAA",
                            "Name": "TestPackage",
                            "ContainerOptions": "Unlocked",
                            "IsDeprecated": False,
                        }
                    ],
                },
                {"size": 0, "records": []},  # No versions
            ]

            Package2 = mock.Mock()
            Package2.update.return_value = {"success": True}
            task._get_tooling_object = mock.Mock(return_value=Package2)

            task._run_task()

            Package2.update.assert_called_once_with(
                "0Ho000000000000AAA", {"IsDeprecated": True}
            )
            mock_confirm.assert_called_once()
